package asd.td4;

import asd.base.Liste;

/**
 * Une classe pour l'exercice 4 de la feuille 4.
 * Lorsqu'elle n'est pas mentionne, la complexit est O(1).
 *  2002 Marc Gatano
 */
public class Ensemble {

    // ATTRIBUTS

    private Liste l; // la liste ordonne de Monome

    // CONSTRUCTEURS

    /**
     * Construit un ensemble vide.
     */
    public Ensemble() {
        l = new Liste();
    }

    /**
     * Construit un singleton contenant i.
     */
    public Ensemble(int i) {
        l = new Liste(new Integer(i));
    }

    /**
     * Construit un ensemble dont le plus petit lment est i
     * et les autres lments sont ceux de l'ensemble e.
     * Prcondition : pour x dans e, i < e.
     */
    private Ensemble(int i, Ensemble e) {
        l = new Liste(new Integer(i),e.l);
    }

    /**
     * Construit un ensemble  partir de la liste l.
     * Prcondition : la liste l est une liste strictement croissante
     * d'entiers.
     */
    private Ensemble(Liste l) {
        this.l = l;
    }

    // PRIMITIVES

    /**
     * Retourne le plus petit entier de l'ensemble.
     * Prcondition : l'ensemble n'est pas vide.
     */
    private int premier() {
        return ((Integer) l.premier()).intValue();
    }

    /**
     * Retourne l'ensmble priv de son plus petit lment.
     * Prcondition : l'ensemble n'est pas vide.
     */
    private Ensemble reste() {
        return new Ensemble(l.reste());
    }

    /**
     * Teste si l'ensemble est vide.
     */
    public boolean vide() {
        return l.vide();
    }

    // NON-PRIMITIVES  CRIRE  l'AIDE DES PRIMITIVES

    /**
     * Retourne le cardinal de l'ensemble.
     * Complexit : O(n) o n est le cardinal de l'ensemble.
     */
    public int cardinal() {
        if ( vide() ) {
            return 0;
        }
        return 1 + reste().cardinal();
    }

    /**
     * Teste si l'ensemble contient l'entier i.
     * Complexit : O(n) o n est le cardinal de l'ensemble.
     */
    public boolean contient(int i) {
        if ( vide() || i < premier() ) {
            return false;
        }
        return i == premier() || reste().contient(i);
    }

    /**
     * Retourne l'union de l'ensemble avec l'ensemble e.
     * Complexit : O(max(n,m)) o n et m sont les cardinaux
     * des deux ensembles.
     */
    public Ensemble union(Ensemble e) {
        if ( vide() ) {
            return e;
        }
        if ( e.vide() ) {
            return this;
        }
        int p = premier();
        int pe = e.premier();
        if ( p < pe ) {
            return new Ensemble(p,reste().union(e));
        }
        if ( p > pe ) {
            return new Ensemble(pe,union(e.reste()));
        }
        return new Ensemble(p,reste().union(e.reste()));
    }

    /**
     * Retourne l'intersection de l'ensemble avec l'ensemble e.
     * Complexit : O(max(n,m)) o n et m sont les cardinaux
     * des deux ensembles.
     */
    public Ensemble inter(Ensemble e) {
        if ( vide() || e.vide() ) {
            return new Ensemble();
        }
        int p = premier();
        int pe = e.premier();
        if ( p < pe ) {
            return reste().inter(e);
        }
        if ( p > pe ) {
            return inter(e.reste());
        }
        return new Ensemble(p,reste().inter(e.reste()));
    }

    /**
     * Retourne la diffrence symtrique de l'ensemble avec l'ensemble e.
     * Complexit : O(max(n,m)) o n et m sont les cardinaux
     * des deux ensembles.
     */
    public Ensemble difference(Ensemble e) {
        if ( vide() ) {
            return new Ensemble();
        }
        if ( e.vide() ) {
            return this;
        }
        int p = premier();
        int pe = e.premier();
        if ( p < pe ) {
            return new Ensemble(p,reste().difference(e));
        }
        if ( p > pe ) {
            return difference(e.reste());
        }
        return reste().difference(e.reste());
    }

    /**
     * Teste si l'ensemble est inclus dans l'ensemble e.
     * Complexit : O(max(n,m)) o n et m sont les cardinaux
     * des deux ensembles.
     */
    public boolean estInclusDans(Ensemble e) {
        if ( vide() ) {
            return true;
        }
        if ( e.vide() ) {
            return false;
        }
        int p = premier();
        int pe = e.premier();
        if ( p < pe ) {
            return false;
        }
        if ( p > pe ) {
            return estInclusDans(e.reste());
        }
        return reste().estInclusDans(e.reste());
    }

    /**
     * Retourne une chane de caractres reprsentant l'ensemble.
     * Complexit : O(n) o n est le cardinal de l'ensemble.
     */
    public String toString() {
        String s = "{";
        Ensemble e  = this;
        if ( ! e.vide() ) {
            s = s + e.premier();
            e = e.reste();
        }
        while ( ! e.vide() ) {
            s = s + ", " + e.premier();
            e = e.reste();
        }
        return s + "}";
    }
}
